#!/usr/bin/env node

const express = require('express')
const morgan = require('morgan')
const program = require('commander')
const package = require('./package.json')

const last = xs => xs.length > 0 ? xs[xs.length - 1] : null

const init = xs => xs.slice(0, xs.length - 1)

const entries = obj => Object.keys(obj).map(k => ([k, obj[k]]))

const ensurePath = path => {
  if (path === '') return null
  if (path.indexOf('./') !== 0 && path.indexOf('/') !== 0) return `${process.cwd()}/${path}`
  else return path
}

const getName = path => init(last(path.split('/')).split('.')).join('-')

const getPath = (name, length, args) => {
  let argsString = ''
  if (args) argsString = args.map(arg => `/:${arg}`).join('')
  else for (let i = 0; i < length; i++) argsString += `/:arg${i}?`
  return '/' + name + argsString
}

const getEndpoints = paths => paths
  .map(ensurePath)
  .filter(path => !!path)
  .map(filepath => {
    const name = getName(filepath)
    const module = require(filepath)
    const fn = module instanceof Function ? module : module.fn
    const args = module ? module.args : null
    const path = getPath(name, fn.length, args)
    return { name, path, fn, args }
  })

const getParams = (args, params) => {
  if (!args) return entries(params).map(([k, v]) => v)
  else return args.map(arg => params[arg.replace('?', '')])
}

const listen = (port, endpoints) => {
  const app = express()

  app.use(morgan('tiny'))

  app.get('/', (req, res) => {
    res.send({ endpoints })
  })

  endpoints.forEach(({ name, path, fn, args }) => {
    app.get(path, (req, res) => {
      let promise = null
      try {
        promise = Promise.resolve(fn.apply(null, getParams(args, req.params)))
      } catch (e) {
        promise = Promise.reject(e)
      }

      promise.then(
        result => res.send(result),
        ({ message, status }) => res.status(status || 500).send(message)
      )
    })
  })

  console.info('Endpoints:')
  console.info(endpoints.map(({ path }) => '  GET ' + path).join('\n'))

  app.listen(port, () => {
    console.info(`Listening on ${port}\n`)
  })
}

program
  .version(package.version)
  .usage('[options] <file ...>')
  .option('-p, --port [port]', 'use the specified port')
  .parse(process.argv)

const port = program.port || process.env.PORT || 3000
const endpoints = getEndpoints(program.args)

if (endpoints.length === 0) program.help()
else listen(port, endpoints)

